import asyncio
import logging
import typing as t
from inspect import isawaitable

from typing_extensions import Self

from ._type import _P, _T, TyImpl

log = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from .static import StaticHookSpec

__all__ = ["HookSpec"]


async def call_impls(
    impls: t.Iterable[TyImpl[_P, _T]], *args: _P.args, **kwds: _P.kwargs
) -> t.AsyncGenerator[_T, t.Any]:
    """A uniform interface to call a batch of hook implements."""
    rfuts: t.List[t.Union[asyncio.Future[_T], _T]] = []
    for impl in impls:
        try:
            c = impl(*args, **kwds)
        except BaseException:
            log.error("sync listener error!", exc_info=True)
            continue

        rfuts.append(asyncio.ensure_future(c) if isawaitable(c) else c)  # type: ignore
        asyncio.Future

    for rfut in rfuts:
        if asyncio.isfuture(rfut):
            try:
                yield await rfut
            except asyncio.CancelledError:
                raise
            except BaseException:
                log.error("async listener error!", exc_info=True)
                continue
        else:
            yield rfut  # type: ignore


class HookSpec(t.Generic[_P, _T]):
    """An instance of a :class:`StaticHookSpec`.

    .. seealso:: :meth:`StaticHookSpec.__call__`
    """

    __slots__ = ("__def__", "impls")

    impls: t.List[TyImpl[_P, _T]]
    """You can access this list to add/remove/clear_all implements."""

    def __init__(self, hookdef: "StaticHookSpec[_P, _T]") -> None:
        super().__init__()
        self.impls = []
        self.__def__ = hookdef.__def__

    def add_impl(self, impl: TyImpl[_P, _T]) -> Self:
        """A shortcut to :obj:`.impls`.append."""
        self.impls.append(impl)
        return self

    @t.overload
    def replace_impl(
        self, index_or_impl: t.Union[int, TyImpl[_P, _T]]
    ) -> t.Callable[[TyImpl[_P, _T]], None]: ...

    @t.overload
    def replace_impl(
        self, index_or_impl: t.Union[int, TyImpl[_P, _T]], new_impl: TyImpl[_P, _T]
    ) -> None: ...

    def replace_impl(
        self,
        index_or_impl: t.Union[int, TyImpl[_P, _T]],
        new_impl: t.Optional[TyImpl[_P, _T]] = None,
    ) -> t.Union[None, t.Callable[[TyImpl[_P, _T]], None]]:
        """Replace an implementation according to the given index or function.

        :param index_or_impl: The index or the function.
            The function will be transformed to index by :obj:`.impls`.index.
        :param new_impl: If not given, this method could be used as a decorator.

        :raise ValueError: If :obj:`index_or_impl` is a function AND not present in :obj:`.impls`.
        """
        if not isinstance(index_or_impl, int):
            index_or_impl = self.impls.index(index_or_impl)

        if new_impl is None:
            return lambda f: self.impls.__setitem__(index_or_impl, f)

        self.impls[index_or_impl] = new_impl

    async def gather(self, *args: _P.args, **kwds: _P.kwargs) -> t.List[_T]:
        """Gather all results, results respsect the corresponding order in :obj:`.impls`.

        :return: All results corresponding the order in :obj:`.impls`
        """
        return [i async for i in call_impls(self.impls, *args, **kwds)]

    results = gather
    """An alias to :meth:`.gather`."""

    async def emit(self, *args: _P.args, **kwds: _P.kwargs) -> None:
        """Like :meth:`.gather`, but returns `None`."""
        await self.gather(*args, **kwds)

    async def first(self, *args: _P.args, **kwds: _P.kwargs) -> _T:
        """
        Get the first valid result.

        If all results are invalid, raises :exc:`StopAsyncIteration`.

        :raise StopAsyncIteration: If all results are invalid.

        :return: The first valid result.
        """
        for impl in self.impls:
            try:
                c = impl(*args, **kwds)
                if isawaitable(c):
                    return await c
                return c  # type: ignore
            except asyncio.CancelledError:
                raise
            except BaseException:
                log.error("listener error", exc_info=True)
                continue

        raise StopAsyncIteration

    async def __call__(self, *args: _P.args, **kwds: _P.kwargs) -> _T:
        """
        Get the first valid result.
        If all results are invalid, call the defination and return its result.

        :return: The first valid result. If all results are invalid, returns the result of the defination.
        """
        try:
            return await self.first(*args, **kwds)
        except StopAsyncIteration:
            c = self.__def__(*args, **kwds)
            return await c if isawaitable(c) else c  # type: ignore

    @property
    def has_impl(self) -> bool:
        """Return if a hook is implemented.

        :return: If this hook is implemented.
        """
        return bool(self.impls)



import asyncio
import typing as t

T = t.TypeVar("T")


class FutureStore:
    """
    A store for keeping references of running futures,
    in order to keep them from being garbage collected.
    """

    _futs: t.Set[asyncio.Future]

    def __init__(self, loop: asyncio.AbstractEventLoop | None = None) -> None:
        """Create a :class:`FutureStore`.

        :param loop: Event loop for running futures. Must be given when using ``thread_safe`` keyword.
        """
        super().__init__()
        self._futs = set()
        self.loop = loop
        """Event loop for running futures. Must be given when using ``thread_safe`` keyword."""

    @t.overload
    def add_awaitable(
        self, func: t.Coroutine[t.Any, t.Any, T], thread_safe: t.Literal[True]
    ) -> "asyncio.Future[T]": ...

    @t.overload
    def add_awaitable(
        self, func: t.Awaitable[T], thread_safe: bool = False
    ) -> "asyncio.Future[T]": ...

    def add_awaitable(
        self, func: t.Awaitable[T], thread_safe: bool = False
    ) -> "asyncio.Future[T]":
        """Add an awaitable into this store.

        :param func: the awaitable
        :param thread_safe: whether to use :meth:`asyncio.run_coroutine_threadsafe`. If true, must specify :obj:`.loop`, and `func` must be a coroutine.
        :return: the wrapped future
        """
        if thread_safe:
            assert self.loop, "`thread_safe` mode needs specify loop when creating FutureStore."
            assert asyncio.iscoroutine(func), "`thread_safe` mode needs coroutine as input."
            func = asyncio.wrap_future(
                asyncio.run_coroutine_threadsafe(func, loop=self.loop), loop=self.loop
            )
        else:
            func = asyncio.ensure_future(func, loop=self.loop)
        if not func.done():
            self._futs.add(func)
            func.add_done_callback(self._futs.discard)
        return func

    __call__ = add_awaitable

    async def wait(self, wait_new: bool = True):
        """Wait for all tasks in the specific group(s).

        :param wait_new: also wait for new-added futures during waiting for existing futures.

        .. note:: Even if `wait_new` is False, newly added tasks may still be scheduled.
        """
        if not self._futs:
            return
        await asyncio.wait(self._futs)
        if wait_new and self._futs:
            # await potential new tasks in this store
            await self.wait()

    def clear(self):
        """Clear current future store, cancel all saved tasks. If you do not want a task to be cancelled,
        wrap it in a `~asyncio.shield`.

        :param exc: if not given, all futures will be cancelled; else they will be set with this exception.
        """
        for task in self._futs:
            task.cancel()

    def __bool__(self):
        return bool(self._futs)



import logging
import typing as t
from functools import WRAPPER_ASSIGNMENTS
from typing import Any

from ._type import _P, _T, TyImpl
from .base import HookSpec
from .timeout import TimeoutHookSpec

__all__ = ["StaticHookSpec"]

log = logging.getLogger(__name__)


class StaticHookSpec(t.Generic[_P, _T]):
    """Defines a hook. The wrapped function will be the original defination."""

    __slots__ = ("__def__",)

    def __init__(self, defination: TyImpl[_P, _T]) -> None:
        self.__def__ = defination

    @property
    def TyInst(self) -> t.Type[HookSpec[_P, _T]]:
        """Designed for type checkers. Used to annotate the type of a derived :class:`HookSpec`."""
        return HookSpec

    @property
    def TyTmInst(self) -> t.Type[TimeoutHookSpec[_P, _T]]:
        """Designed for type checkers. Used to annotate the type of a derived :class:`TimeoutHookSpec`."""
        return TimeoutHookSpec

    def __getattribute__(self, __name: str) -> Any:
        if __name in WRAPPER_ASSIGNMENTS:
            return getattr(self.__def__, __name)
        if __name == "__annotations__":
            # python 3.14: __annotations__ lazy evaluation
            # BUG: is this readonly?
            return getattr(self.__def__, __name, {})
        return super().__getattribute__(__name)

    def __call__(self) -> HookSpec[_P, _T]:
        """Get a :class:`HookSpec` of this hook."""
        return HookSpec(self)

    def with_timeout(self, timeout: float):
        """Get a :class:`TimeoutHookSpec` of this hook."""
        return TimeoutHookSpec(self, timeout=timeout)



import asyncio
import typing as t
from contextlib import suppress
from functools import wraps

from ._type import _P, _T
from .base import HookSpec

if t.TYPE_CHECKING:
    from .static import StaticHookSpec


__all__ = ["TimeoutHookSpec"]


class TimeoutHookSpec(HookSpec[_P, _T]):
    """A HookSpec with a pre-defined timeout.

    Internally, :meth:`.gather` and :meth:`.first` is wrapped with :external+python:obj:`asyncio.wait_for`.
    Thus a :exc:`asyncio.TimeoutError` will be raised if timeout.

    .. seealso::

        :meth:`tylisten.StaticHookSpec.with_timeout`
        :external+python:obj:`asyncio.wait_for`
    """

    def __init__(self, hookdef: "StaticHookSpec[_P, _T]", *, timeout: float) -> None:
        super().__init__(hookdef)
        self._timeout = timeout

        timeout_meths = [self.gather, self.first]
        for meth in timeout_meths:
            self.__with_timeout(meth.__name__)

    def __with_timeout(self, name: str):
        meth = getattr(self, name)

        @wraps(meth)
        async def wrapper(*args, **kwds):
            return await asyncio.wait_for(meth(*args, **kwds), timeout=self._timeout)

        setattr(self, name, wrapper)
        return meth

    async def emit_suppress_timeout(self, *args: _P.args, **kwds: _P.kwargs) -> None:
        """Call :meth:`.emit` and suppress :exc:`asyncio.TimeoutError`"""
        with suppress(asyncio.TimeoutError):
            return await super().emit(*args, **kwds)
